/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Histogram - JGL - 18/06/2012
// 
// Juan Gómez-Luna*, José María González-Linares#, José Ignacio Benavides* and Nicolás Guil#
// *University of Córdoba, #University of Málaga (Spain)
// http://www.ac.uma.es/~vip/
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// RGB to grayscale conversion and histogram by applying R-per-block approach to histogram computation
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// histogram_RGB2gray.cu
// This file contains the CPU code for grayscale histogram computation of RGB images.
// It launches the R-per-block kernel.
//
// Reference:
// @article {springerlink:10.1007/s00138-012-0443-3,
//   author = {Gómez-Luna, Juan and González-Linares, José and Benavides, José and Guil, Nicolás},
//   affiliation = {Department of Computer Architecture and Electronics, University of Córdoba, Córdoba, Spain},
//   title = {An optimized approach to histogram computation on GPU},
//   journal = {Machine Vision and Applications},
//   publisher = {Springer Berlin / Heidelberg},
//   issn = {0932-8092},
//   keyword = {Computer Science},
//   pages = {1-10},
//   url = {http://dx.doi.org/10.1007/s00138-012-0443-3},
//   note = {10.1007/s00138-012-0443-3},
//}
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <cutil.h>

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Reference CPU histogram
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
extern "C" void histo_CPU(unsigned int *histo, unsigned int *data, int size);

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// GPU kernels
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#include <histogram_RGB2gray_kernel.cu>

//Carry out dummy calculations before main computation loop
//in order to "warm up" the hardware/driver
#define WARMUP

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Main program
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv){
	const int DATA_W = 768; // 2560;
	const int DATA_H = 576; // 1920;
	unsigned char *h_DataA;
	unsigned int *h_Lect;
	unsigned int *hh_DataA; 
	unsigned int *d_DataA; 
	unsigned int *d_Temp;

	unsigned int *d_histo_block;

	unsigned int *h_ResultCPU, *h_ResultGPU_block;

	float time_histo_block, time_block;
	time_histo_block = 0; time_block = 0;
	unsigned int hTimer;
	double time_cpu = 0;
	double sum_delta, sum_ref, L1norm;

	cutCreateTimer(&hTimer);

	// Syntax verification
	if (argc != 7) {
		printf("Wrong format\n");
		printf("Syntax: %s <Device BINS NUMBER_OF_BLOCKS THREADS_PER_BLOCK REPLICATION_FACTOR REPEAT>\n",argv[0]);
		exit(1);
	}

	// Check the compute capability of the device
	int num_devices=0;
	cudaGetDeviceCount(&num_devices);
	if(0==num_devices){
		printf("Your system does not have a CUDA capable device\n");
		return 1;
	}
	int device = atoi(argv[1]);
	cudaDeviceProp device_properties;
	cudaGetDeviceProperties(&device_properties,device);
	if((1==device_properties.major)&&(device_properties.minor<2))
		printf("%s does not have compute capability 1.2 or later\n\n",device_properties.name);

	int BINS = atoi(argv[2]);
	int DATA_SIZE = DATA_W * DATA_H;
	int NUM_BLOCKS = atoi(argv[3]);
	int THREADS = atoi(argv[4]);
	int REP = atoi(argv[5]);
	int repeat = atoi(argv[6]);

	int BINSp = BINS + 1;

	const int DATA_SIZE_INT = DATA_SIZE * sizeof(unsigned int);

	printf("%s on %s\n", argv[0], device_properties.name);
	printf("%d repetitions\n", repeat);
	printf("Image size = %d pixels\n", DATA_SIZE);
	printf("Execution configuration: %d blocks of %d threads, %d sub-histograms per block\n", NUM_BLOCKS, THREADS, REP);
	printf("-------------------------------------------------------------------------------\n");

	// Host memory allocation
	cudaMallocHost((void **)&h_DataA, DATA_SIZE_INT);
	cudaMallocHost((void **)&h_Lect, DATA_SIZE_INT*3);
	cudaMallocHost((void **)&hh_DataA, DATA_SIZE_INT);

    cudaMallocHost((void **)&h_ResultCPU, BINS*sizeof(unsigned int));
    cudaMallocHost((void **)&h_ResultGPU_block, BINS*sizeof(unsigned int));

	// Device memory allocation
	cudaMalloc((void **)&d_DataA, DATA_SIZE_INT);

    cudaMalloc((void **)&d_Temp, BINS*sizeof(unsigned int));
    cudaMalloc((void **)&d_histo_block, BINS*sizeof(unsigned int));

	// Initizalize host and device arrays
	memset(h_ResultCPU,0,BINS*sizeof(unsigned int));
	memset(h_ResultGPU_block,0,BINS*sizeof(unsigned int));
	cudaMemset(d_histo_block,0,BINS*sizeof(unsigned int));

	///////////////////////////////////////////////////////////////////////////////////////////////////
	// Grids and blocks
	///////////////////////////////////////////////////////////////////////////////////////////////////
	dim3 blockGrid_histo(NUM_BLOCKS);
	dim3 threadBlock_histo(THREADS);

	///////////////////////////////////////////////////////////////////////////////////////////////////
#ifdef WARMUP
	// RANDOM WORKLOAD
	srand(time(NULL));
	for(int i=0; i<DATA_SIZE * 4; ++i){ // Pixels are stored in 4 bytes (RGBa)
		h_DataA[i] = (unsigned char)(BINS*(((float)rand()/(float)RAND_MAX)));
	}
	// Host to device memory
	cudaMemcpy(d_DataA, h_DataA, DATA_SIZE_INT, cudaMemcpyHostToDevice);
	//printf("Warm up...\n");
	histo_R_per_block<<<blockGrid_histo,threadBlock_histo,(BINSp*REP)*sizeof(int)>>>(
		d_Temp,d_DataA,DATA_SIZE,BINS,BINSp,REP);
	cudaThreadSynchronize();
#endif
	///////////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////
	cudaEvent_t start, stop;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);

    char dctFileName[100];
    FILE *File;
	
for(int i = 0; i < repeat; i++){
	///////////////////////////////////////////////////////////////////////////////////////////////////
	// Input image, from file
	///////////////////////////////////////////////////////////////////////////////////////////////////
	memset(h_Lect,0,DATA_SIZE_INT*3);
	sprintf(dctFileName, "image_McGill.rgb");
	//printf("%s\n", dctFileName);
	if ((File = fopen(dctFileName,"rt")) != NULL){
		for (int y=0; y<DATA_SIZE*3; ++y){
			int sc = fscanf(File,"%u ",h_Lect+y);
       	}
    	fclose(File);
    }
	memset(h_DataA,0,DATA_SIZE_INT); int k = 0;
    for(int l=0; l<DATA_SIZE * 4; l=l+4){ // Pixels are stored in 4 bytes (RGBa)
		for(int j=0; j<3; ++j){
	        h_DataA[l+j] = (unsigned char)(h_Lect[k]);
			k++;
		}
    }

	// Host to device memory
	cudaMemcpy(d_DataA, h_DataA, DATA_SIZE_INT, cudaMemcpyHostToDevice);

	memset(h_ResultCPU,0,BINS*sizeof(unsigned int));
	cudaMemset(d_histo_block,0,BINS*sizeof(unsigned int));

	///////////////////////////////////////////////////////////////////////////////////////////////////
	// R-PER-BLOCK APPROACH - IN SHARED MEMORY
	///////////////////////////////////////////////////////////////////////////////////////////////////
	cudaEventRecord( start, 0 );

	histo_R_per_block<<<blockGrid_histo,threadBlock_histo,(BINSp*REP)*sizeof(int)>>>(
		d_histo_block,d_DataA,DATA_SIZE,BINS,BINSp,REP);

	cudaEventRecord( stop, 0 );
	cudaEventSynchronize( stop );
	cudaEventElapsedTime( &time_histo_block, start, stop );
	time_block += time_histo_block;

	///////////////////////////////////////////////////////////////////////////////////////////////////
	// CPU HISTOGRAM
	///////////////////////////////////////////////////////////////////////////////////////////////////
	CUDA_SAFE_CALL(cudaMemcpy(hh_DataA, h_DataA, DATA_SIZE_INT, cudaMemcpyHostToHost));
	// Start CPU timer
	CUDA_SAFE_CALL(cudaThreadSynchronize());	// Thread synchronization
	CUT_SAFE_CALL(cutResetTimer(hTimer));
	CUT_SAFE_CALL(cutStartTimer(hTimer));

	histo_CPU(h_ResultCPU, hh_DataA, DATA_SIZE);

	// Stop CPU timer
	CUDA_SAFE_CALL(cudaThreadSynchronize());	// Thread synchronization
	CUT_SAFE_CALL(cutStopTimer(hTimer));
	time_cpu += cutGetTimerValue(hTimer);

	/*printf("\n");
	for(int y=0; y<BINS; ++y){
		printf("%d ",h_ResultCPU[y]);
	}*/
	CUDA_SAFE_CALL(cudaMemcpy(h_ResultGPU_block,d_histo_block,BINS*sizeof(int),cudaMemcpyDeviceToHost));
	/*printf("\n");
	for(int y=0; y<BINS; ++y){
		printf("%d ",h_ResultGPU_block[y]);
	}*/
}

printf("R_per_block approach on GPU =\t%.4f\tms\n",time_block/repeat);

printf("CPU computation =\t%.4f\tms\n",time_cpu/repeat);

printf("Speedup =\t%.2f\n", (time_cpu/repeat)/(time_block/repeat));

	///////////////////////////////////////////////////////////////////////////////////////////////////
	// Comparing the results
	///////////////////////////////////////////////////////////////////////////////////////////////////
	sum_delta = 0;
	sum_ref   = 0;
	L1norm = 0;
	for(int i = 0; i < BINS; i++){
		sum_delta += fabs(h_ResultGPU_block[i] - h_ResultCPU[i]);
    	sum_ref   += fabs(h_ResultCPU[i]);
	}
	L1norm = (double)(sum_delta / sum_ref);
	printf((L1norm < 1e-6) ? "TEST PASSED\n" : "TEST FAILED\n");

	///////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	///////////////////////////////////////////////////////////////////////////////////////////////////
	// Device memory
	cudaFree(d_Temp);
    cudaFree(d_DataA);
	cudaFree(d_histo_block);
	// Host memory
	cudaFreeHost(h_DataA);
	cudaFreeHost(h_Lect);
	cudaFreeHost(hh_DataA);
	cudaFreeHost(h_ResultGPU_block);
	cudaFreeHost(h_ResultCPU);

	CUT_SAFE_CALL(cutDeleteTimer(hTimer));
	cudaEventDestroy(start);
    cudaEventDestroy(stop);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
