/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Histogram - JGL - 18/06/2012
// 
// Juan Gómez-Luna*, José María González-Linares#, José Ignacio Benavides* and Nicolás Guil#
// *University of Córdoba, #University of Málaga (Spain)
// http://www.ac.uma.es/~vip/
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Histogram of monochrome image by applying R-per-block approach to histogram computation
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// histogram_VanHateren_kernel.cu
// This file contains the GPU version of histogram computation of monochrome images by applying R-per-block approach.
//
// Reference:
// @article {springerlink:10.1007/s00138-012-0443-3,
//   author = {Gómez-Luna, Juan and González-Linares, José and Benavides, José and Guil, Nicolás},
//   affiliation = {Department of Computer Architecture and Electronics, University of Córdoba, Córdoba, Spain},
//   title = {An optimized approach to histogram computation on GPU},
//   journal = {Machine Vision and Applications},
//   publisher = {Springer Berlin / Heidelberg},
//   issn = {0932-8092},
//   keyword = {Computer Science},
//   pages = {1-10},
//   url = {http://dx.doi.org/10.1007/s00138-012-0443-3},
//   note = {10.1007/s00138-012-0443-3},
//}
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Constants
#define WARP_SIZE 32

// Dynamic shared memory allocation
extern __shared__ unsigned int Hs[];

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// R-per-block approach
// Replication + Padding + Interleaved read access
//
// histo:	Final histogram in global memory
// data:	Input image. Pixels are stored in 4-byte unsigned int
// size:	Input image size (number of pixels)
// BINS:	Number of histogram bins
// BINSp:	Histogram size including padding
// R:		Replication factor
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
__global__ void histo_R_per_block(unsigned int* histo, unsigned int* data, int size, int BINS, int BINSp, int R){
	// Block and thread index
	const int bx = blockIdx.x;
	const int tx = threadIdx.x;
	// Warp and lane
	const unsigned int warpid = tx >> 5;
	const unsigned int lane = tx & 31;	

	// Offset to per-block sub-histograms
	const unsigned int off_rep = BINSp * (tx % R);

	// Constants for interleaved read access
	const int warps_block = blockDim.x / WARP_SIZE;
	const int begin = (size / warps_block) * warpid + WARP_SIZE * bx + lane;
	const int end = (size / warps_block) * (warpid + 1);
	const int step = WARP_SIZE * gridDim.x;
	// Constants for naive read access
	/*const int begin = bx * blockDim.x + tx;
	const int end = size;
	const int step = blockDim.x * gridDim.x;*/

	// Sub-histograms initialization
	for(int pos = tx; pos < BINSp*R; pos += blockDim.x) Hs[pos] = 0;

	__syncthreads();	// Intra-block synchronization

	// Main loop
	for(int i = begin; i < end; i += step){
		// Global memory read
		unsigned int d = data[i];

		// Atomic vote in shared memory
		atomicAdd(&Hs[off_rep + ((d * BINS) >> 12)], 1);

	}

	__syncthreads();	// Intra-block synchronization

	// Merge per-block histograms and write to global memory
	for(int pos = tx; pos < BINS; pos += blockDim.x){
		unsigned int sum = 0;
		for(int base = 0; base < BINSp*R; base += BINSp)
			sum += Hs[base + pos];
		// Atomic addition in global memory
		atomicAdd(histo + pos, sum);
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
